from abc import ABC, abstractmethod
from collections import UserDict

import xmltodict


class SATParser(ABC):
    @abstractmethod
    def parse(self, response: str) -> UserDict:
        """Parse an event from a source in XML representation."""


class VerifyParser(SATParser):
    @classmethod
    def parse(cls, response: str) -> UserDict:
        """Gets the Query ID from the raw response"""
        response_dict = xmltodict.parse(response)
        result = response_dict["Envelope"]["Body"]["VerificaSolicitudDescargaResponse"][
            "VerificaSolicitudDescargaResult"
        ]
        data = {
            "EstadoSolicitud": result["@EstadoSolicitud"],
            "CodEstatus": result["@CodEstatus"],
            "Mensaje": result["@Mensaje"],
            "CodigoEstadoSolicitud": result["@CodigoEstadoSolicitud"],
            "NumeroCFDIs": result["@NumeroCFDIs"],
            "IdsPaquetes": [result["IdsPaquetes"]]
            if result["@EstadoSolicitud"] == "3"
            else "",  # TODO Check what happens when multiple ids
        }
        return data


class QueryParser(SATParser):
    @classmethod
    def parse(cls, response: str) -> UserDict:
        """Gets the Query ID from the raw response"""
        response_dict = xmltodict.parse(response)
        result = response_dict["Envelope"]["Body"]["SolicitaDescargaResponse"][
            "SolicitaDescargaResult"
        ]
        data = {
            "CodEstatus": result["@CodEstatus"],
            "IdSolicitud": result.get("@IdSolicitud"),
        }
        return data


class DownloadParser(SATParser):
    @classmethod
    def parse(cls, response: str) -> UserDict:
        """Gets the Download data from the raw response"""
        response_dict = xmltodict.parse(response)
        package = response_dict["Envelope"]["Body"]["RespuestaDescargaMasivaTercerosSalida"][
            "Paquete"
        ]
        data = {
            "Content": package,
        }
        return data


class LoginParser(SATParser):
    @classmethod
    def parse(cls, response: str) -> UserDict:
        """Gets the token from the raw response"""
        response_dict = xmltodict.parse(response)
        data = {
            "token": response_dict["Envelope"]["Body"]["AutenticaResponse"]["AutenticaResult"],
        }
        return data



import logging
import urllib

from OpenSSL import crypto

from . import utils

_logger = logging.getLogger(__name__)


class NoUniqueIdentifierException(Exception):
    """If not valid RFC founded in the certificate"""


class NoIssuerException(Exception):
    """If not valid Issuer founded in the certificate"""


class CertificateHandler:
    cert: str
    key: str
    password: str

    unique_identifier: str
    certificate: crypto.X509
    key_pem: str
    cert_pem: str

    def __init__(self, cert_binary: bytes, key_binary: bytes, password: str):
        self.cert = utils.binary_to_utf8(cert_binary)
        self.key = utils.binary_to_utf8(key_binary)
        self.password = password
        self._load_certs()
        self._compute_data_from_cert()

    def _load_certs(self):
        """Loads the PEM version of the certificate and key file, also loads the crypto certificate

        Convert the `cert` and `key` from DER to PEM and creates the real certificate (X509)
        """
        self.key_pem = utils.der_to_pem(self.key, cert_type="ENCRYPTED PRIVATE KEY")
        self.cert_pem = utils.der_to_pem(self.cert, cert_type="CERTIFICATE")
        self.certificate = crypto.load_certificate(crypto.FILETYPE_PEM, self.cert_pem)

    def _compute_data_from_cert(self):
        """Gets the RFC and Issuer directly from the certificate"""
        self._get_rfc_from_cert()
        self._get_issuer_from_cert()

    def _get_rfc_from_cert(self):
        """Gets the RFC from the certificate

        Raises:
            NoUniqueIdentifierException: If not RFC founded
        """
        subject_components = self.certificate.get_subject().get_components()
        for c in subject_components:
            if c[0] == b"x500UniqueIdentifier":
                self.unique_identifier = c[1].decode("UTF-8").split(" ")[0]
                _logger.debug("x500UniqueIdentifier %s loaded", self.unique_identifier)
                break
        else:
            raise NoUniqueIdentifierException()

    def _get_issuer_from_cert(self):
        """Gets the Issuer from the certificate

        Raises:
            NoIssuerException: If not Issuer founded
        """
        self.certificate.issuer = ",".join(
            [
                f'{c[0].decode("UTF-8")}={urllib.parse.quote(c[1].decode("UTF-8"))}'
                for c in self.certificate.get_issuer().get_components()
            ]
        )
        if not self.certificate.issuer:
            raise NoIssuerException()
        _logger.debug("Issuer %s loaded", self.certificate.issuer)

    def sign(self, data: str) -> str:
        """Signs the `data` using SHA1 with the `key_pem` content"""
        _logger.debug("Signing %s", data)
        private_key = crypto.load_privatekey(
            crypto.FILETYPE_PEM, self.key_pem, passphrase=self.password
        )
        signed_data = crypto.sign(private_key, data, "sha1")
        decoded_data = utils.binary_to_utf8(signed_data)
        return decoded_data



import logging
import time
from datetime import datetime

from . import exceptions, utils
from .certificate_handler import CertificateHandler
from .enums import TEMPLATES
from .sat_login_handler import SATLoginHandler
from .sat_parsers import DownloadParser, QueryParser, VerifyParser

logging.basicConfig(level=logging.INFO)
_logger = logging.getLogger(__name__)


class SATConnector:
    """Class to make a connection to the SAT"""

    certificate_handler: CertificateHandler
    login_handler: SATLoginHandler

    def __init__(self, cert: bytes, key: bytes, password: str) -> None:
        """Loads the certificate, key file and password to stablish the connection to the SAT

        Creates a object to manage the SAT connection.

        Args:
            cert (bytes): DER Certificate in raw binary
            key (bytes): DER Key Certificate in raw binary
            password (str): Key password in plain text (utf-8)
        """
        self.certificate_handler = CertificateHandler(cert, key, password)
        self.login_handler = SATLoginHandler(self.certificate_handler)
        _logger.info("Data correctly loaded")

    def _create_common_envelope(self, template: str, data: dict) -> str:
        _logger.debug("Creating Envelope")
        _logger.debug("%s", template)
        _logger.debug("%s", data)
        query_data, query_data_signature = utils.prepare_template(template, data)
        digest_value = utils.digest(query_data)
        signed_info = utils.prepare_template(
            TEMPLATES["SignedInfo"],
            {
                "uri": "",
                "digest_value": digest_value,
            },
        )
        key_info = utils.prepare_template(
            TEMPLATES["KeyInfo"],
            {
                "issuer_name": self.certificate_handler.certificate.issuer,
                "serial_number": self.certificate_handler.certificate.get_serial_number(),
                "certificate": self.certificate_handler.cert,
            },
        )
        signature_value = self.certificate_handler.sign(signed_info)
        signature = utils.prepare_template(
            TEMPLATES["Signature"],
            {
                "signed_info": signed_info,
                "signature_value": signature_value,
                "key_info": key_info,
            },
        )
        envelope_content = utils.prepare_template(
            query_data_signature,
            {
                "signature": signature,
            },
        )
        envelope = utils.prepare_template(
            TEMPLATES["Envelope"],
            {
                "content": envelope_content,
            },
        )
        _logger.debug("Final Envelope")
        _logger.debug("%s", envelope)
        return envelope

    def query(self, start: datetime, end: datetime, download_type: str, request_type: str) -> str:
        """Creates a Query in the SAT system"""
        request_content = self._get_query_soap_body(start, end, download_type, request_type)
        response = utils.consume(
            "http://DescargaMasivaTerceros.sat.gob.mx/ISolicitaDescargaService/SolicitaDescarga",
            "https://cfdidescargamasivasolicitud.clouda.sat.gob.mx/SolicitaDescargaService.svc",
            request_content,
            token=self.login_handler.token,
        )
        if response.status_code != 200:
            raise exceptions.RequestException(
                response.status_code, response.reason, request_content
            )
        response_clean = utils.remove_namespaces(response.content.decode("UTF-8"))
        query_id = QueryParser.parse(response_clean)
        return query_id

    def _get_query_soap_body(
        self, start: datetime, end: datetime, download_type: str, request_type: str
    ):
        """Creates the SOAP body to the query request"""
        start = start.isoformat()
        end = end.isoformat()
        data = {
            "start": start,
            "end": end,
            "rfc": self.certificate_handler.unique_identifier,
            "download_type": download_type,
            "request_type": request_type,
            "signature": "",
        }
        envelope = self._create_common_envelope(TEMPLATES["SolicitaDescarga"], data)
        return envelope

    def verify(self, query_id: str) -> dict:
        """Checks the status of a Query"""
        request_content = self._get_verify_soap_body(query_id)
        response = utils.consume(
            "http://DescargaMasivaTerceros.sat.gob.mx/IVerificaSolicitudDescargaService/VerificaSolicitudDescarga",
            "https://cfdidescargamasivasolicitud.clouda.sat.gob.mx/VerificaSolicitudDescargaService.svc",
            request_content,
            token=self.login_handler.token,
        )
        if response.status_code != 200:
            raise exceptions.RequestException(
                response.status_code, response.reason, request_content
            )
        response_clean = utils.remove_namespaces(response.content.decode("UTF-8"))
        data = VerifyParser.parse(response_clean)
        return data

    def _get_verify_soap_body(self, query_id: str) -> str:
        """Creates the SOAP body to check the query status"""
        data = {
            "rfc": self.certificate_handler.unique_identifier,
            "query_id": query_id,
            "signature": "",
        }
        envelope = self._create_common_envelope(TEMPLATES["VerificaSolicitudDescarga"], data)
        return envelope

    def download(self, package_ids: (list, str)) -> dict:
        """Checks the status of a Query"""
        if isinstance(package_ids, str):
            package_ids = [package_ids]
        downloads = {}
        for package_id in package_ids:
            request_content = self._get_download_soap_body(package_id)
            response = utils.consume(
                "http://DescargaMasivaTerceros.sat.gob.mx/IDescargaMasivaTercerosService/Descargar",
                "https://cfdidescargamasiva.clouda.sat.gob.mx/DescargaMasivaService.svc",
                request_content,
                token=self.login_handler.token,
            )
            if response.status_code != 200:
                raise exceptions.RequestException(
                    response.status_code, response.reason, request_content
                )
            response_clean = utils.remove_namespaces(response.content.decode("UTF-8"))
            downloads[package_id] = DownloadParser.parse(response_clean)
        return downloads

    def _get_download_soap_body(self, package_id: str) -> dict:
        """Creates the SOAP body to check the query status"""
        data = {
            "rfc": self.certificate_handler.unique_identifier,
            "package_id": package_id,
            "signature": "",
        }
        envelope = self._create_common_envelope(
            TEMPLATES["PeticionDescargaMasivaTercerosEntrada"], data
        )
        return envelope

    def wait_query(self, query_id: str, retries: int = 10, wait_seconds: int = 2) -> list:
        for _ in range(retries):
            verification = self.verify(query_id)
            query_state = int(verification["EstadoSolicitud"])
            if query_state > 3:
                raise exceptions.QueryException(f"EstadoSolicitud({query_state})")
            if query_state == 3:
                return verification["IdsPaquetes"]
            time.sleep(wait_seconds)
        raise TimeoutError("The query is not yet resolved")



def test_query(query):
    assert query["CodEstatus"] == "5000"


def test_verify(sat_connector, query):
    response = sat_connector.verify(query["IdSolicitud"])
    return response


def test_download(sat_connector, package_ids):
    downloads = sat_connector.download(package_ids)
    return downloads



import base64
import hashlib
import logging
import re
import textwrap

import requests

logging.basicConfig(level=logging.INFO)
_logger = logging.getLogger(__name__)


def clean_xml(xml: str) -> str:
    """Clean a XML string to be used in SAT request.

    Removes all the spaces and new line characters between tags.

    Args:
        xml (str): XML to be cleaned.

    Returns:
        str: XML clean.
    """
    return re.sub(r"\s+(?=[<>])", "", xml).strip()


def remove_namespaces(xml):
    return re.sub(r"[souh]:", "", xml)


def prepare_template(template: str, data: dict) -> str:  # TODO simplify
    """Takes a XML template and fill the `variable` (data betwen {}) fields.

    Args:
        template (str): Template to be processed.
        data (dict): Variables to be replaced.

    Returns:
        str: Template with variables replaced.
    """
    template_clean = clean_xml(template)
    final_template = template_clean.format(**data)
    if "signature" in data.keys() and not data.get("signature"):
        data["signature"] = "{signature}"
        template_signature_to_replace = template_clean.format(**data)
        return (final_template, template_signature_to_replace)
    return final_template


def binary_to_utf8(binary: bytes) -> str:
    """Takes a bytes object an returns the string represents it.

    Args:
        binary (bytes): Raw binary to be process.

    Returns:
        str: binary in base64 in utf-8.
    """
    return base64.encodebytes(binary).decode("UTF-8")


def digest(data: str) -> str:
    return binary_to_utf8(hashlib.sha1(data.encode("UTF-8")).digest())


def der_to_pem(der_data: str, cert_type: str) -> str:
    """Convert DER data into PEM.

    Args:
        der_data (str): DER data to be convert.
        cert_type (str): Type of certificate to be created
                         (`ENCRYPTED PRIVATE KEY`, `CERTIFICATE`, etc).

    Returns:
        str: Certificate converted.
    """
    wrapped = "\n".join(textwrap.wrap(der_data, 64))
    pem = f"-----BEGIN {cert_type}-----\n{wrapped}\n-----END {cert_type}-----\n"
    return pem


def consume(soap_action, uri, body, token=None):
    headers = {
        "Content-type": 'text/xml; charset="utf-8"',
        "Accept": "text/xml",
        "Cache-Control": "no-cache",
        "SOAPAction": soap_action,
    }
    if token:
        headers["Authorization"] = f'WRAP access_token="{token}"'
    response = requests.post(uri, body, headers=headers)
    return response
